"""Color-related functions and data structures."""

from .color_types import AColor, AHexcode, Color, ColorTheme, Hexcode
from .colors import (
    ABLACK,
    ABLUE,
    ACYAN,
    AGREEN,
    AMAGENTA,
    ARED,
    AWHITE,
    AYELLOW,
    BLACK,
    BLUE,
    CYAN,
    GREEN,
    MAGENTA,
    NEPTUNE_PRIMARY_BG,
    NEPTUNE_PRIMARY_FG,
    NEPTUNE_THEME,
    RED,
    TRANSPARENT,
    WHITE,
    YELLOW,
    Neptune,
)
from .gradients import (
    darken_only,
    gradient,
    lerp_colors,
    lighten_only,
    rainbow_gradient,
)

__all__ = [
    "ABLACK",
    "ABLUE",
    "ACYAN",
    "AGREEN",
    "AMAGENTA",
    "ARED",
    "AWHITE",
    "AYELLOW",
    "BLACK",
    "BLACK_ON_BLACK",
    "BLUE",
    "CYAN",
    "NEPTUNE_THEME",
    "NEPTUNE_PRIMARY_BG",
    "NEPTUNE_PRIMARY_FG",
    "GREEN",
    "MAGENTA",
    "RED",
    "TRANSPARENT",
    "WHITE",
    "WHITE_ON_BLACK",
    "YELLOW",
    "AColor",
    "AHexcode",
    "Color",
    "ColorTheme",
    "Hexcode",
    "Neptune",
    "darken_only",
    "gradient",
    "lerp_colors",
    "lighten_only",
    "rainbow_gradient",
]
